from __future__ import unicode_literals

import io
import logging
import os
import sys
from importlib import import_module

from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from django.core.management.color import no_style
from django.core.management.sql import sql_flush
from django.db import DEFAULT_DB_ALIAS
from django.db import connections
from django.db import transaction
from django.utils import six
from django.utils.six.moves import input

from django_north.management.migrations import get_current_version
from django_north.management.migrations import get_fixtures_for_init
from django_north.management.migrations import fixtures_default_tpl
from django_north.management.runner import Script

logger = logging.getLogger(__name__)


# warning: backport from django 1.8
# this command changed a lot in 1.10


class Command(BaseCommand):
    help = ('Removes ALL DATA from the database, including data added during '
            'migrations. Unmigrated apps will also have their initial_data '
            'fixture reloaded. Does not achieve a "fresh install" state.')

    def add_arguments(self, parser):
        parser.add_argument(
            '--noinput', action='store_false', dest='interactive',
            default=True,
            help='Tells Django to NOT prompt the user for input of any kind.')
        parser.add_argument(
            '--database', action='store', dest='database',
            default=DEFAULT_DB_ALIAS,
            help='Nominates a database to flush. Defaults to the "default" '
            'database.')
        parser.add_argument(
            '--no-initial-data', action='store_false',
            dest='load_initial_data', default=True,
            help='Tells Django not to load any initial data after database '
            'synchronization.')

    def handle(self, **options):
        if getattr(settings, 'NORTH_MANAGE_DB', False) is not True:
            logger.info('flush command disabled')
            return

        self.flush(**options)

    def flush(self, **options):
        database = options.get('database')
        connection = connections[database]
        verbosity = options.get('verbosity')
        interactive = options.get('interactive')
        # The following are stealth options used by Django's internals.
        reset_sequences = options.get('reset_sequences', True)
        allow_cascade = options.get('allow_cascade', False)
        inhibit_post_migrate = options.get('inhibit_post_migrate', False)

        self.style = no_style()

        # Import the 'management' module within each installed app, to register
        # dispatcher events.
        for app_config in apps.get_app_configs():
            try:
                import_module('.management', app_config.name)
            except ImportError:
                pass

        # custom: only_django False
        # get current version before flush
        current_version = get_current_version()
        sql_list = sql_flush(self.style, connection, only_django=False,
                             reset_sequences=reset_sequences,
                             allow_cascade=allow_cascade)

        if interactive:
            confirm = input(
                """You have requested a flush of the database.
This will IRREVERSIBLY DESTROY all data currently in the %r database,
and return each table to an empty state.
Are you sure you want to do this?

    Type 'yes' to continue, or 'no' to cancel: """
                % connection.settings_dict['NAME'])
        else:
            confirm = 'yes'

        if confirm == 'yes':
            try:
                with transaction.atomic(
                        using=database,
                        savepoint=connection.features.can_rollback_ddl):
                    with connection.cursor() as cursor:
                        for sql in sql_list:
                            cursor.execute(sql)
            except Exception as e:
                new_msg = (
                    "Database %s couldn't be flushed. Possible reasons:\n"
                    "  * The database isn't running or isn't configured "
                    "correctly.\n"
                    "  * At least one of the expected database tables doesn't "
                    "exist.\n"
                    "  * The SQL was invalid.\n"
                    "Hint: Look at the output of 'django-admin sqlflush'. "
                    "That's the SQL this command wasn't able to run.\n"
                    "The full error: %s") % (
                        connection.settings_dict['NAME'], e)
                six.reraise(
                    CommandError, CommandError(new_msg), sys.exc_info()[2])

            if not inhibit_post_migrate:
                self.emit_post_migrate(
                    verbosity, interactive, database, current_version)

            # Reinstall the initial_data fixture.
            if options.get('load_initial_data'):
                # Reinstall the initial_data fixture for apps without
                # migrations.
                from django.db.migrations.executor import MigrationExecutor
                executor = MigrationExecutor(connection)
                app_options = options.copy()
                for app_label in executor.loader.unmigrated_apps:
                    app_options['app_label'] = app_label
                    try:
                        call_command('loaddata', 'initial_data', **app_options)
                    except CommandError:
                        # fails with django 1.10 if initial_data does not exist
                        pass
        else:
            self.stdout.write("Flush cancelled.\n")

    @staticmethod
    def emit_post_migrate(verbosity, interactive, database, current_version):
        # custom: do what was done on post_migrate
        # clear contenttype cache
        ContentType.objects.clear_cache()

        # reload fixtures
        connection = connections[database]
        fixtures_version = get_fixtures_for_init(current_version)
        fixtures_path = os.path.join(
            settings.NORTH_MIGRATIONS_ROOT,
            'fixtures',
            getattr(settings, 'NORTH_FIXTURES_TPL', fixtures_default_tpl)
            .format(fixtures_version))
        with io.open(fixtures_path, 'r', encoding='utf8') as f:
            script = Script(f)
            script.run(connection)



# -*- coding: utf-8 -*-

from django.apps import apps
from django.core.management.base import BaseCommand

from django_north.management import contenttypes
from django_north.management import permissions


class Command(BaseCommand):
    help = "Displays all the missing fixtures."

    def add_arguments(self, parser):
        parser.add_argument(
            "--unknown-contenttypes", action="store_true",
            dest="unknown_contenttypes", default=False,
            help="Show unknown contenttypes to be removed")

    def handle(self, *args, **options):
        fixtures = []
        if options['unknown_contenttypes']:
            fixtures += self.unknown_contenttypes()
        fixtures += self.missing_contenttypes()
        fixtures += self.missing_permissions()

        return "\n".join(fixtures) + "\n"

    def unknown_contenttypes(self):
        """
        Contenttypes to remove
        """
        cts = []
        for app_config in apps.get_app_configs():
            cts += contenttypes.get_unknown_contenttypes_for_app_config(
                app_config)
        return [
            "DELETE FROM django_content_type "
            "WHERE app_label = '{}' AND model = '{}';".format(
                ct.app_label, ct.model)
            for ct in cts
        ]

    def missing_contenttypes(self):
        """
        Contenttypes to create
        """
        cts = []
        for app_config in apps.get_app_configs():
            cts += contenttypes.get_missing_contenttypes_for_app_config(
                app_config)
        return [
            "INSERT INTO django_content_type(app_label, model) "
            "VALUES('{}', '{}');".format(ct.app_label, ct.model)
            for ct in cts
        ]

    def missing_permissions(self):
        """
        Permissions to create
        """
        perms = []
        for app_config in apps.get_app_configs():
            perms += permissions.get_missing_permissions_for_app_config(
                app_config)
        return [
            "INSERT INTO auth_permission(codename, name, content_type_id) "
            "VALUES('{}', '{}', (SELECT id FROM django_content_type "
            "WHERE app_label = '{}' AND model = '{}'));".format(
                perm.codename, perm.name, perm.content_type.app_label,
                perm.content_type.model)
            for perm in perms
        ]



from django.contrib.contenttypes.models import ContentType
from django.db import DEFAULT_DB_ALIAS
from django.utils import six


def get_all_contenttypes_for_app_config(app_config, using=DEFAULT_DB_ALIAS):
    """
    Return all existing contenttypes for a given app_config
    """
    app_label = app_config.label

    return {
        ct.model: ct
        for ct in ContentType.objects.using(using).filter(app_label=app_label)
    }


def get_known_models_for_app_config(app_config):
    """
    Return all known models for a given app config
    """
    return {
        model._meta.model_name: model
        for model in app_config.get_models()}


def get_unknown_contenttypes_for_app_config(app_config):
    """
    Return unknown contenttypes for a given app_config
    """
    app_models = get_known_models_for_app_config(app_config)
    if not app_models:
        return []

    content_types = get_all_contenttypes_for_app_config(app_config)
    if not content_types:
        return []

    return [
        ct
        for (model_name, ct) in six.iteritems(content_types)
        if model_name not in app_models
    ]


def get_missing_contenttypes_for_app_config(app_config):
    """
    Return missing contenttypes for a given app_config
    """
    app_label = app_config.label

    app_models = get_known_models_for_app_config(app_config)
    if not app_models:
        return []

    content_types = get_all_contenttypes_for_app_config(app_config)

    return [
        ContentType(
            app_label=app_label,
            model=model_name,
        )
        for (model_name, model) in six.iteritems(app_models)
        if model_name not in content_types
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')
version = open('VERSION').read().strip()

setup(
    name='django-north',
    version=version,
    description="""Yet another way to manage migrations: DBA as a service""",
    long_description=readme + '\n\n' + history,
    author=u'Lauréline Guérin',
    author_email='laureline.guerin@people-doc.com',
    url='https://github.com/novafloss/django-north',
    packages=[
        'django_north',
    ],
    include_package_data=True,
    install_requires=[
        "Django>=1.8,<2.0",
        "sqlparse",
        "six",
    ],
    tests_require=["tox"],
    license="MIT",
    zip_safe=False,
    keywords='django-north',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



import io
import os
from distutils.version import StrictVersion
from importlib import import_module

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import connection
from django.db.migrations.recorder import \
    MigrationRecorder as DjangoMigrationRecorder
from django.db.utils import ProgrammingError


fixtures_default_tpl = 'fixtures_{}.sql'
schema_default_tpl = 'schema_{}.sql'


class DBException(Exception):
    pass


def is_version(vstring):
    try:
        StrictVersion(vstring)
    except ValueError:
        return False
    return True


def list_dirs(root):
    return [d for d in os.listdir(root)
            if os.path.isdir(os.path.join(root, d))]


def list_files(root):
    return [d for d in os.listdir(root)
            if os.path.isfile(os.path.join(root, d))]


class MigrationRecorder(DjangoMigrationRecorder):
    def record_applied(self, app, name):
        """
        Records that a migration was applied.
        """
        # custom: do not ensure schema
        self.migration_qs.create(app=app, name=name)


def get_known_versions():
    """
    Return the list of the known versions defined in migration repository,
    ordered.
    Ignore symlinks.
    """
    # get all subfolders
    try:
        dirs = list_dirs(settings.NORTH_MIGRATIONS_ROOT)
    except OSError:
        raise ImproperlyConfigured(
            'settings.NORTH_MIGRATIONS_ROOT is improperly configured.')

    # exclude symlinks and some folders (like schemas, fixtures, etc)
    versions = [
        d for d in dirs
        if not os.path.islink(os.path.join(settings.NORTH_MIGRATIONS_ROOT, d))
        and is_version(d)]

    # sort versions
    versions.sort(key=StrictVersion)
    return versions


def get_applied_versions():
    """
    Return the list of applied versions.
    Reuse django migration table.
    """
    recorder = MigrationRecorder(connection)
    applied_versions = list(recorder.migration_qs.filter(
        app__in=get_known_versions()
    ).values_list(
        'app', flat=True).order_by('app').distinct())

    # sort versions
    applied_versions.sort(key=StrictVersion)
    return applied_versions


def get_current_version():
    """
    Return the current version of the database.
    Return None if the schema is not inited.
    """
    try:
        import_string = settings.NORTH_CURRENT_VERSION_DETECTOR
    except AttributeError:
        import_string = (
            'django_north.management.migrations'
            '.get_current_version_from_table')

    module_path, factory_name = import_string.rsplit('.', 1)
    module = import_module(module_path)
    factory = getattr(module, factory_name)

    return factory()


def get_current_version_from_table():
    """
    Return the current version of the database, from sql_version table.
    Return None if the table does not exist (schema not inited).
    """
    with connection.cursor() as cursor:
        try:
            cursor.execute("SELECT version_num FROM sql_version;")
        except ProgrammingError:
            # table does not exist ?
            return None

        rows = cursor.fetchall()

    versions = [row[0] for row in rows if is_version(row[0])]
    if not versions:
        return None

    # sort versions
    versions.sort(key=StrictVersion)

    # return the last one
    return versions[-1]


def get_current_version_from_comment():
    """
    Return the current version of the database, from django_site comment.
    Return None if the table django_site does not exist (schema not inited).
    """
    with connection.cursor() as cursor:
        try:
            cursor.execute(
                "SELECT obj_description('django_site'::regclass, 'pg_class');")
        except ProgrammingError:
            # table does not exist ?
            return None

        row = cursor.fetchone()
        comment = row[0]

    # no comment
    if comment is None:
        raise DBException('No comment found on django_site.')

    # parse comment
    if 'version ' not in comment:
        raise DBException("No version found in django_site's comment.")
    return comment.replace('version ', '').strip()


def get_applied_migrations(version):
    """
    Return the list of applied migrations for the given version.
    Reuse django migration table.
    """
    recorder = MigrationRecorder(connection)
    return list(recorder.migration_qs.filter(app=version).values_list(
        'name', flat=True))


def get_migrations_to_apply(version):
    """
    Return an dict containing the list of migrations to apply for
    the given version.
    Key: name of the migration.
    Value: path of the migration.
    """
    def filter_migrations(files):
        return [
            f for f in files
            if f.endswith('ddl.sql') or f.endswith('dml.sql')]

    version_root = os.path.join(settings.NORTH_MIGRATIONS_ROOT, version)
    migrations = {}

    # list auto migrations
    try:
        files = list_files(version_root)
    except OSError:
        raise DBException('No sql folder found for version {}.'.format(
            version))
    # filter files (keep *ddl.sql and *dml.sql)
    auto_migrations = filter_migrations(files)
    # store migrations
    for mig in auto_migrations:
        migrations[mig] = os.path.join(version_root, mig)

    # list manual migrations
    manual_root = os.path.join(version_root, 'manual')
    try:
        files = list_files(manual_root)
    except OSError:
        files = []
    # filter files (keep *ddl.sql and *dml.sql)
    auto_migrations = filter_migrations(files)
    # store migrations
    for mig in auto_migrations:
        migrations[mig] = os.path.join(manual_root, mig)

    return migrations


def get_closest_version(target_version, sql_tpl):
    """
    Get the version of a file (schema or fixtures) to use to init a DB.
    Take the closest to the target_version. Can be the same version, or older.
    """
    # get known versions
    known_versions = get_known_versions()
    # find target version
    try:
        target_version_index = known_versions.index(target_version)
    except ValueError:
        raise ImproperlyConfigured(
            'settings.NORTH_TARGET_VERSION is improperly configured: '
            'version {} not found.'.format(
                settings.NORTH_TARGET_VERSION))

    def get_version(index):
        version = known_versions[index]
        schema_path = sql_tpl.format(version)
        return version, schema_path

    while target_version_index > 0:
        version, path = get_version(target_version_index)
        if os.path.exists(path):
            return version
        target_version_index -= 1

    version, path = get_version(target_version_index)
    if os.path.exists(path):
        return version


def get_version_for_init():
    """
    Get the version to use to init a new DB to the current target version.
    """
    version = get_closest_version(
        settings.NORTH_TARGET_VERSION,
        os.path.join(
            settings.NORTH_MIGRATIONS_ROOT,
            'schemas',
            getattr(settings, 'NORTH_SCHEMA_TPL', schema_default_tpl)))

    if version is None:
        raise DBException('Can not find a schema to init the DB.')
    return version


def get_fixtures_for_init(target_version):
    """
    Get the closest fixtures to use to init a new DB
    to the current target version.
    """
    version = get_closest_version(
        target_version,
        os.path.join(
            settings.NORTH_MIGRATIONS_ROOT,
            'fixtures',
            getattr(settings, 'NORTH_FIXTURES_TPL', fixtures_default_tpl)))

    if version is None:
        raise DBException('Can not find fixtures to init the DB.')
    return version


def is_manual_migration(file_handler):
    if '/manual/' in file_handler.name:
        return True

    if not file_handler.name.endswith('dml.sql'):
        return False

    for line in file_handler:
        if '--meta-psql:' in line:
            file_handler.seek(0)
            return True

    file_handler.seek(0)
    return False


def build_migration_plan():
    """
    Return the list of migrations by version,
    from the version used to init the DB to the current target version.
    """
    # get current version
    current_version = get_current_version()
    if current_version is None:
        # schema not inited
        return None
    # get known versions
    known_versions = get_known_versions()
    # get applied versions
    applied_versions = get_applied_versions()

    migration_plan = {
        'current_version': current_version,
        'init_version': None,
        'plans': [],
    }

    # guess the version used to init the DB
    if not applied_versions:
        # current version is not None, it was used to init the DB
        try:
            first_version_index = known_versions.index(current_version) + 1
        except ValueError:
            raise DBException(
                'The current version of the database is unknown: {}.'
                .format(current_version))
        init_version = current_version
    else:
        first_version_index = known_versions.index(applied_versions[0])
        init_version = known_versions[first_version_index - 1]
    migration_plan['init_version'] = init_version

    # get all versions to apply
    try:
        target_version_index = known_versions.index(
            settings.NORTH_TARGET_VERSION)
    except ValueError:
        raise ImproperlyConfigured(
            'settings.NORTH_TARGET_VERSION is improperly configured: '
            'version {} not found.'.format(
                settings.NORTH_TARGET_VERSION))
    versions_to_apply = known_versions[
        first_version_index:target_version_index + 1]

    # get plan for each version to apply
    for version in versions_to_apply:
        version_plan = []
        # get applied migrations
        applied_migrations = get_applied_migrations(version)
        # get migrations to apply
        migrations_to_apply = get_migrations_to_apply(version)
        migs = list(migrations_to_apply.keys())
        migs.sort()
        # build plan
        for mig in migs:
            applied = mig in applied_migrations
            path = migrations_to_apply[mig]
            with io.open(path, 'r', encoding='utf8') as f:
                is_manual = is_manual_migration(f)
            version_plan.append((mig, applied, path, is_manual))
        migration_plan['plans'].append({
            'version': version,
            'plan': version_plan
        })

    return migration_plan



# -*- coding: utf-8 -*-
import logging

from django.conf import settings
from django.core.management.base import BaseCommand

from django_north.management import migrations

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Shows all available migrations for the current project"

    def handle(self, *args, **options):
        if getattr(settings, 'NORTH_MANAGE_DB', False) is not True:
            logger.info('showmigrations command disabled')
            return

        return self.show_list()

    def show_list(self):
        """
        Shows a list of all migrations on the system,
        from the version used to init the DB, to the current target version.
        """
        migration_plan = migrations.build_migration_plan()
        if migration_plan is None:
            self.stdout.write(self.style.ERROR("Schema not inited"))
            return

        self.stdout.write(
            self.style.MIGRATE_HEADING("Current version of the DB:"))
        self.stdout.write(
            self.style.MIGRATE_LABEL(
                "  {}".format(migration_plan['current_version'])))

        self.stdout.write(
            self.style.MIGRATE_HEADING("Schema used to init the DB:"))
        self.stdout.write(
            self.style.MIGRATE_LABEL(
                "  {}".format(migration_plan['init_version'])))

        # display migration status for each version to apply
        for plan in migration_plan['plans']:
            self.stdout.write(self.style.MIGRATE_HEADING("Version:"))
            self.stdout.write(
                self.style.MIGRATE_LABEL("  {}".format(plan['version'])))
            # print plan
            for mig, applied, path, is_manual in plan['plan']:
                title = mig
                if is_manual:
                    title += ' (manual)'
                if applied:
                    self.stdout.write("  [X] %s" % title)
                else:
                    self.stdout.write("  [ ] %s" % title)
